import functools
import glob
import locale
import sys
from contextlib import contextmanager
from datetime import timedelta
from os.path import expanduser, isdir

import click
import click_log

from todoman import exceptions, formatters
from todoman.configuration import ConfigurationException, load_config
from todoman.interactive import TodoEditor
from todoman.model import cached_property, Database, Todo


click_log.basic_config()


@contextmanager
def handle_error():
    try:
        yield
    except exceptions.TodomanException as e:
        click.echo(e)
        sys.exit(e.EXIT_CODE)


def catch_errors(f):
    @functools.wraps(f)
    def wrapper(*a, **kw):
        with handle_error():
            return f(*a, **kw)

    return wrapper


TODO_ID_MIN = 1
with_id_arg = click.argument('id', type=click.IntRange(min=TODO_ID_MIN))


def _validate_lists_param(ctx, param=None, lists=()):
    return [_validate_list_param(ctx, name=l) for l in lists]


def _validate_list_param(ctx, param=None, name=None):
    ctx = ctx.find_object(AppContext)
    if name is None:
        if ctx.config['main']['default_list']:
            name = ctx.config['main']['default_list']
        else:
            raise click.BadParameter(
                'You must set "default_list" or use -l.'.format(name)
            )
    lists = {
        list_.name: list_
        for list_ in ctx.db.lists()
    }
    fuzzy_matches = [
        list_
        for list_ in lists.values()
        if list_.name.lower() == name.lower()
    ]

    if len(fuzzy_matches) == 1:
        return fuzzy_matches[0]

    # case-insensitive matching collides or does not find a result,
    # use exact matching
    if name in lists:
        return lists[name]
    raise click.BadParameter(
        "{}. Available lists are: {}".format(
            name, ', '.join(list_.name for list_ in lists.values()))
    )


def _validate_date_param(ctx, param, val):
    ctx = ctx.find_object(AppContext)
    try:
        return ctx.formatter.parse_datetime(val)
    except ValueError as e:
        raise click.BadParameter(e)


def _validate_priority_param(ctx, param, val):
    ctx = ctx.find_object(AppContext)
    try:
        return ctx.formatter.parse_priority(val)
    except ValueError as e:
        raise click.BadParameter(e)


def _validate_start_date_param(ctx, param, val):
    ctx = ctx.find_object(AppContext)
    if not val:
        return val

    if len(val) != 2 or val[0] not in ['before', 'after']:
        raise click.BadParameter("Format should be '[before|after] [DATE]'")

    is_before = val[0] == 'before'

    try:
        dt = ctx.formatter.parse_datetime(val[1])
        return is_before, dt
    except ValueError as e:
        raise click.BadParameter(e)


def _validate_startable_param(ctx, param, val):
    ctx = ctx.find_object(AppContext)
    return val or ctx.config['main']['startable']


def _validate_todos(ctx, param, val):
    ctx = ctx.find_object(AppContext)
    with handle_error():
        return [ctx.db.todo(int(id)) for id in val]


def _sort_callback(ctx, param, val):
    fields = val.split(',') if val else []
    for field in fields:
        if field.startswith('-'):
            field = field[1:]

        if field not in Todo.ALL_SUPPORTED_FIELDS and field != 'id':
            raise click.BadParameter('Unknown field "{}"'.format(field))

    return fields


def validate_status(ctx=None, param=None, val=None):
    # The default command doesn't run callbacks as expected, so it needs to
    # specify the callback'd type. When `list` is called explicitly, this
    # callback *IS* run, so we need to handle that edge case:
    if not isinstance(val, str):
        return val

    statuses = val.upper().split(',')

    if 'ANY' in statuses:
        return Todo.VALID_STATUSES

    for status in statuses:
        if status not in Todo.VALID_STATUSES:
            raise click.BadParameter(
                'Invalid status, "{}", statuses must be one of "{}", or "ANY"'
                .format(status, ', '.join(Todo.VALID_STATUSES))
            )

    return statuses


def _todo_property_options(command):
    click.option(
        '--priority',
        default='',
        callback=_validate_priority_param,
        help=('The priority for this todo')
    )(command)
    click.option(
        '--location', help=('The location where '
                            'this todo takes place.')
    )(command)
    click.option(
        '--due',
        '-d',
        default='',
        callback=_validate_date_param,
        help=(
            'The due date of the task, in the format specified in the '
            'configuration file.'
        )
    )(command)
    click.option(
        '--start',
        '-s',
        default='',
        callback=_validate_date_param,
        help='When the task starts.'
    )(command)

    @functools.wraps(command)
    def command_wrap(*a, **kw):
        kw['todo_properties'] = {
            key: kw.pop(key)
            for key in ('due', 'start', 'location', 'priority')
        }
        return command(*a, **kw)

    return command_wrap


class AppContext:
    def __init__(self):
        self.config = None
        self.db = None
        self.formatter_class = None

    @cached_property
    def ui_formatter(self):
        return formatters.DefaultFormatter(
            self.config['main']['date_format'],
            self.config['main']['time_format'],
            self.config['main']['dt_separator']
        )

    @cached_property
    def formatter(self):
        return self.formatter_class(
            self.config['main']['date_format'],
            self.config['main']['time_format'],
            self.config['main']['dt_separator']
        )


pass_ctx = click.make_pass_decorator(AppContext)

_interactive_option = click.option(
    '--interactive',
    '-i',
    is_flag=True,
    default=None,
    help='Go into interactive mode before saving the task.'
)


@click.group(invoke_without_command=True)
@click_log.simple_verbosity_option()
@click.option(
    '--colour',
    '--color',
    'colour',
    default=None,
    type=click.Choice(['always', 'auto', 'never']),
    help=(
        'By default todoman will disable colored output if stdout '
        'is not a TTY (value `auto`). Set to `never` to disable '
        'colored output entirely, or `always` to enable it '
        'regardless.'
    )
)
@click.option(
    '--porcelain',
    is_flag=True,
    help='Use a JSON format that will '
    'remain stable regardless of configuration or version.'
)
@click.option(
    '--humanize',
    '-h',
    default=None,
    is_flag=True,
    help='Format all dates and times in a human friendly way'
)
@click.option(
    '--config',
    '-c',
    default=None,
    help='The config file to use.',
    envvar='TODOMAN_CONFIG',
    metavar='PATH',
)
@click.pass_context
@click.version_option(prog_name='todoman')
@catch_errors
def cli(click_ctx, colour, porcelain, humanize, config):
    ctx = click_ctx.ensure_object(AppContext)
    try:
        ctx.config = load_config(config)
    except ConfigurationException as e:
        raise click.ClickException(e.args[0])

    if porcelain and humanize:
        raise click.ClickException(
            '--porcelain and --humanize cannot be used'
            ' at the same time.'
        )

    if humanize is None:  # False means explicitly disabled
        humanize = ctx.config['main']['humanize']

    if porcelain:
        ctx.formatter_class = formatters.PorcelainFormatter
    elif humanize:
        ctx.formatter_class = formatters.HumanizedFormatter
    else:
        ctx.formatter_class = formatters.DefaultFormatter

    colour = colour or ctx.config['main']['color']
    if colour == 'always':
        click_ctx.color = True
    elif colour == 'never':
        click_ctx.color = False

    paths = [
        path for path in glob.iglob(expanduser(ctx.config["main"]["path"]))
        if isdir(path)
    ]
    if len(paths) == 0:
        raise exceptions.NoListsFound(ctx.config["main"]["path"])

    ctx.db = Database(paths, ctx.config['main']['cache_path'])

    # Make python actually use LC_TIME, or the user's locale settings
    locale.setlocale(locale.LC_TIME, "")

    if not click_ctx.invoked_subcommand:
        invoke_command(
            click_ctx,
            ctx.config['main']['default_command'],
        )


def invoke_command(click_ctx, command):
    name, *args = command.split(' ')
    if name not in cli.commands:
        raise click.ClickException(
            'Invalid setting for [main][default_command]'
        )
    click_ctx.invoke(cli.commands[command], args)


try:  # pragma: no cover
    import click_repl
    click_repl.register_repl(cli)
    click_repl.register_repl(cli, name="shell")
except ImportError:
    pass


@cli.command()
@click.argument('summary', nargs=-1)
@click.option(
    '--list',
    '-l',
    callback=_validate_list_param,
    help='The list to create the task in.'
)
@click.option(
    '--read-description',
    '-r',
    is_flag=True,
    default=False,
    help='Read task description from stdin.'
)
@_todo_property_options
@_interactive_option
@pass_ctx
@catch_errors
def new(ctx, summary, list, todo_properties, read_description, interactive):
    '''
    Create a new task with SUMMARY.
    '''

    todo = Todo(new=True, list=list)

    default_due = ctx.config['main']['default_due']
    if default_due:
        todo.due = todo.created_at + timedelta(hours=default_due)

    default_priority = ctx.config['main']['default_priority']
    if default_priority is not None:
        todo.priority = default_priority

    for key, value in todo_properties.items():
        if value:
            setattr(todo, key, value)
    todo.summary = ' '.join(summary)

    if read_description:
        todo.description = '\n'.join(sys.stdin)

    if interactive or (not summary and interactive is None):
        ui = TodoEditor(todo, ctx.db.lists(), ctx.ui_formatter)
        ui.edit()
        click.echo()  # work around lines going missing after urwid

    if not todo.summary:
        raise click.UsageError('No SUMMARY specified')

    ctx.db.save(todo)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.option(
    '--raw',
    is_flag=True,
    help=(
        'Open the raw file for editing in $EDITOR.\n'
        "Only use this if you REALLY know what you're doing!"
    )
)
@_todo_property_options
@_interactive_option
@with_id_arg
@catch_errors
def edit(ctx, id, todo_properties, interactive, raw):
    '''
    Edit the task with id ID.
    '''
    todo = ctx.db.todo(id)
    if raw:
        click.edit(filename=todo.path)
        return
    old_list = todo.list

    changes = False
    for key, value in todo_properties.items():
        if value:
            changes = True
            setattr(todo, key, value)

    if interactive or (not changes and interactive is None):
        ui = TodoEditor(todo, ctx.db.lists(), ctx.ui_formatter)
        ui.edit()

    # This little dance avoids duplicates when changing the list:
    new_list = todo.list
    todo.list = old_list
    ctx.db.save(todo)
    if old_list != new_list:
        ctx.db.move(todo, new_list=new_list, from_list=old_list)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@with_id_arg
@catch_errors
def show(ctx, id):
    '''
    Show details about a task.
    '''
    todo = ctx.db.todo(id, read_only=True)
    click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.argument(
    'todos',
    nargs=-1,
    required=True,
    type=click.IntRange(0),
    callback=_validate_todos,
)
@catch_errors
def done(ctx, todos):
    """Mark one or more tasks as done."""
    for todo in todos:
        todo.complete()
        ctx.db.save(todo)
        click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.argument(
    'todos',
    nargs=-1,
    required=True,
    type=click.IntRange(0),
    callback=_validate_todos,
)
@catch_errors
def cancel(ctx, todos):
    """Cancel one or more tasks."""
    for todo in todos:
        todo.cancel()
        ctx.db.save(todo)
        click.echo(ctx.formatter.detailed(todo))


@cli.command()
@pass_ctx
@click.confirmation_option(
    prompt='Are you sure you want to delete all done tasks?'
)
@catch_errors
def flush(ctx):
    '''
    Delete done tasks. This will also clear the cache to reset task IDs.
    '''
    database = ctx.db
    for todo in database.flush():
        click.echo(ctx.formatter.simple_action('Flushing', todo))


@cli.command()
@pass_ctx
@click.argument('ids', nargs=-1, required=True, type=click.IntRange(0))
@click.option('--yes', is_flag=True, default=False)
@catch_errors
def delete(ctx, ids, yes):
    '''
    Delete tasks.

    Permanently deletes one or more task. It is recommended that you use the
    `cancel` command if you wish to remove this from the pending list, but keep
    the actual task around.
    '''

    todos = []
    for i in ids:
        todo = ctx.db.todo(i)
        click.echo(ctx.formatter.compact(todo))
        todos.append(todo)

    if not yes:
        click.confirm('Do you want to delete those tasks?', abort=True)

    for todo in todos:
        click.echo(ctx.formatter.simple_action('Deleting', todo))
        ctx.db.delete(todo)


@cli.command()
@pass_ctx
@click.option(
    '--list',
    '-l',
    callback=_validate_list_param,
    help='The list to copy the tasks to.'
)
@click.argument('ids', nargs=-1, required=True, type=click.IntRange(0))
@catch_errors
def copy(ctx, list, ids):
    '''Copy tasks to another list.'''

    for id in ids:
        original = ctx.db.todo(id)
        todo = original.clone()
        todo.list = list
        click.echo(ctx.formatter.compact(todo))
        ctx.db.save(todo)


@cli.command()
@pass_ctx
@click.option(
    '--list',
    '-l',
    callback=_validate_list_param,
    help='The list to move the tasks to.'
)
@click.argument('ids', nargs=-1, required=True, type=click.IntRange(0))
@catch_errors
def move(ctx, list, ids):
    '''Move tasks to another list.'''

    for id in ids:
        todo = ctx.db.todo(id)
        click.echo(ctx.formatter.compact(todo))
        ctx.db.move(todo, list)


@cli.command()
@pass_ctx
@click.argument('lists', nargs=-1, callback=_validate_lists_param)
@click.option('--location', help='Only show tasks with location containg TEXT')
@click.option('--category', help='Only show tasks with category containg TEXT')
@click.option('--grep', help='Only show tasks with message containg TEXT')
@click.option(
    '--sort',
    help=(
          'Sort tasks using fields like : '
          '"start", "due", "priority", "created_at", "percent_complete" etc.'
          '\nFor all fields please refer to: '
          '<https://todoman.readthedocs.io/en/stable/usage.html> '
          ),
    callback=_sort_callback,
)
@click.option(
    '--reverse/--no-reverse',
    default=True,
    help='Sort tasks in reverse order (see --sort). '
    'Defaults to true.'
)
@click.option(
    '--due',
    default=None,
    help='Only show tasks due in INTEGER '
    'hours',
    type=int
)
@click.option(
    '--priority',
    default=None,
    help='Only show tasks with'
    ' priority at least as high as TEXT (low, medium or high).',
    type=str,
    callback=_validate_priority_param
)
@click.option(
    '--start',
    default=None,
    callback=_validate_start_date_param,
    nargs=2,
    help='Only shows tasks before/after given DATE'
)
@click.option(
    '--startable',
    default=None,
    is_flag=True,
    callback=_validate_startable_param,
    help='Show only todos which '
    'should can be started today (i.e.: start time is not in the '
    'future).'
)
@click.option(
    '--status',
    '-s',
    default=['NEEDS-ACTION', 'IN-PROCESS'],
    callback=validate_status,
    help='Show only todos with the '
    'provided comma-separated statuses. Valid statuses are '
    '"NEEDS-ACTION", "CANCELLED", "COMPLETED", "IN-PROCESS" or "ANY"'
)
@catch_errors
def list(ctx, *args, **kwargs):
    """
    List tasks. Filters any completed or cancelled tasks by default.

    If no arguments are provided, all lists will be displayed, and only
    incomplete tasks are show. Otherwise, only todos for the specified list
    will be displayed.

    eg:
      \b
      - `todo list' shows all unfinished tasks from all lists.
      - `todo list work' shows all unfinished tasks from the list `work`.

    This is the default action when running `todo'.

    The following commands can further filter shown todos, or include those
    omited by default:
    """
    todos = ctx.db.todos(**kwargs)
    click.echo(ctx.formatter.compact_multiple(todos))



from datetime import date, datetime, timedelta
from unittest.mock import patch

import pytest
import pytz
from dateutil.tz import tzlocal
from dateutil.tz.tz import tzoffset
from freezegun import freeze_time

from todoman.exceptions import AlreadyExists
from todoman.model import cached_property, Database, List, Todo


def test_querying(create, tmpdir):
    for list in 'abc':
        for i, location in enumerate('abc'):
            create(
                'test{}.ics'.format(i),
                ('SUMMARY:test_querying\r\n'
                 'LOCATION:{}\r\n').format(location),
                list_name=list
            )

    db = Database([str(tmpdir.ensure_dir(l)) for l in 'abc'],
                  str(tmpdir.join('cache')))

    assert len(set(db.todos())) == 9
    assert len(set(db.todos(lists='ab'))) == 6
    assert len(set(db.todos(lists='ab', location='a'))) == 2


def test_retain_tz(tmpdir, create, todos):
    create(
        'ar.ics', 'SUMMARY:blah.ar\n'
        'DUE;VALUE=DATE-TIME;TZID=HST:20160102T000000\n'
    )
    create(
        'de.ics', 'SUMMARY:blah.de\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20160102T000000\n'
    )

    todos = list(todos())

    assert len(todos) == 2
    assert todos[0].due == datetime(
        2016, 1, 2, 0, 0, tzinfo=tzoffset(None, -36000)
    )
    assert todos[1].due == datetime(
        2016, 1, 2, 0, 0, tzinfo=tzoffset(None, 3600)
    )


def test_due_date(tmpdir, create, todos):
    create('ar.ics', 'SUMMARY:blah.ar\n' 'DUE;VALUE=DATE:20170617\n')

    todos = list(todos())

    assert len(todos) == 1
    assert todos[0].due == date(2017, 6, 17)


def test_change_paths(tmpdir, create):
    old_todos = set('abcdefghijk')
    for x in old_todos:
        create('{}.ics'.format(x), 'SUMMARY:{}\n'.format(x), x)

    tmpdir.mkdir('3')

    db = Database([tmpdir.join(x) for x in old_todos],
                  tmpdir.join('cache.sqlite'))

    assert {t.summary for t in db.todos()} == old_todos

    db.paths = [str(tmpdir.join('3'))]
    db.update_cache()

    assert len(list(db.lists())) == 1
    assert not list(db.todos())


def test_list_displayname(tmpdir):
    tmpdir.join('default').mkdir()
    with tmpdir.join('default').join('displayname').open('w') as f:
        f.write('personal')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.name == 'personal'
    assert str(list_) == 'personal'


def test_list_colour(tmpdir):
    tmpdir.join('default').mkdir()
    with tmpdir.join('default').join('color').open('w') as f:
        f.write('#8ab6d2')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour == '#8ab6d2'


def test_list_colour_cache_invalidation(tmpdir, sleep):
    tmpdir.join('default').mkdir()
    with tmpdir.join('default').join('color').open('w') as f:
        f.write('#8ab6d2')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour == '#8ab6d2'

    sleep()

    with tmpdir.join('default').join('color').open('w') as f:
        f.write('#f874fd')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour == '#f874fd'


def test_list_no_colour(tmpdir):
    tmpdir.join('default').mkdir()

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour is None


def test_database_priority_sorting(create, todos):
    for i in [1, 5, 9, 0]:
        create('test{}.ics'.format(i), 'PRIORITY:{}\n'.format(i))
    create('test_none.ics'.format(i), 'SUMMARY:No priority (eg: None)\n')

    todos = list(todos())

    assert todos[0].priority == 0
    assert todos[1].priority == 0
    assert todos[2].priority == 9
    assert todos[3].priority == 5
    assert todos[4].priority == 1


def test_retain_unknown_fields(tmpdir, create, default_database):
    """
    Test that we retain unknown fields after a load/save cycle.
    """
    create(
        'test.ics', 'UID:AVERYUNIQUEID\n'
        'SUMMARY:RAWR\n'
        'X-RAWR-TYPE:Reptar\n'
    )

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite'))
    todo = db.todo(1, read_only=False)

    todo.description = 'Rawr means "I love you" in dinosaur.'
    default_database.save(todo)

    path = tmpdir.join('default').join('test.ics')
    with path.open() as f:
        vtodo = f.read()
    lines = vtodo.splitlines()

    assert 'SUMMARY:RAWR' in lines
    assert 'DESCRIPTION:Rawr means "I love you" in dinosaur.' in lines
    assert 'X-RAWR-TYPE:Reptar' in lines


def test_todo_setters(todo_factory):
    todo = todo_factory()

    todo.description = 'A tea would be nice, thanks.'
    assert todo.description == 'A tea would be nice, thanks.'

    todo.priority = 7
    assert todo.priority == 7

    now = datetime.now()
    todo.due = now
    assert todo.due == now

    todo.description = None
    assert todo.description == ''

    todo.priority = None
    assert todo.priority == 0

    todo.categories = None
    assert todo.categories == []

    todo.due = None
    assert todo.due is None


@freeze_time('2017-03-19-15')
def test_is_completed():
    completed_at = datetime(2017, 3, 19, 14, tzinfo=pytz.UTC)

    todo = Todo()
    assert todo.is_completed is False

    todo.completed_at = completed_at
    assert todo.is_completed is True

    todo.percent_complete = 20
    todo.complete()
    assert todo.is_completed is True
    assert todo.completed_at == datetime.now(pytz.UTC)
    assert todo.percent_complete == 100
    assert todo.status == 'COMPLETED'


@pytest.mark.parametrize(
    'until',
    [
        '20990315T020000Z',  # TZ-aware UNTIL
        '20990315T020000',  # TZ-naive UNTIL
    ]
)
@pytest.mark.parametrize(
    'tz',
    [
        pytz.UTC,  # TZ-aware todos
        None,  # TZ-naive todos
    ]
)
@pytest.mark.parametrize('due', [
    True,
    False,
])
def test_complete_recurring(default_database, due, todo_factory, tz, until):
    # We'll lose the milis when casting, so:
    now = datetime.now(tz).replace(microsecond=0)

    if bool(tz) != bool(until.endswith('Z')):
        pytest.skip('These combinations are invalid, as per the spec.')

    original_start = now
    if due:
        original_due = now + timedelta(hours=12)
    else:
        due = original_due = None

    rrule = 'FREQ=DAILY;UNTIL={}'.format(until)
    todo = todo_factory(rrule=rrule, due=original_due, start=original_start)

    todo.complete()
    related = todo.related[0]

    if due:
        assert todo.due == original_due
    else:
        assert todo.due is None
    assert todo.start == original_start
    assert todo.is_completed
    assert not todo.rrule

    if due:
        assert related.due == original_due + timedelta(days=1)
    else:
        assert related.due is None
    assert related.start == original_start + timedelta(days=1)
    # check due/start tz
    assert not related.is_completed
    assert related.rrule == rrule


def test_save_recurring_related(default_database, todo_factory, todos):
    now = datetime.now(pytz.UTC)
    original_due = now + timedelta(hours=12)
    rrule = 'FREQ=DAILY;UNTIL=20990315T020000Z'
    todo = todo_factory(rrule=rrule, due=original_due)
    todo.complete()

    default_database.save(todo)

    todos = todos(status='ANY')
    todo = next(todos)
    assert todo.percent_complete == 100
    assert todo.is_completed is True
    assert not todo.rrule

    todo = next(todos)
    assert todo.percent_complete == 0
    assert todo.is_completed is False
    assert todo.rrule == rrule


def test_todo_filename_absolute_path():
    Todo(filename='test.ics')
    with pytest.raises(ValueError):
        Todo(filename='/test.ics')


def test_list_equality(tmpdir):
    list1 = List(path=str(tmpdir), name='test list')
    list2 = List(path=str(tmpdir), name='test list')
    list3 = List(path=str(tmpdir), name='yet another test list')

    assert list1 == list2
    assert list1 != list3
    assert list1 != 'test list'


def test_clone():
    now = datetime.now(tz=tzlocal())

    todo = Todo(new=True)
    todo.summary = 'Organize a party'
    todo.location = 'Home'
    todo.due = now
    todo.uid = '123'
    todo.id = '123'
    todo.filename = '123.ics'

    clone = todo.clone()

    assert todo.summary == clone.summary
    assert todo.location == clone.location
    assert todo.due == clone.due
    assert todo.uid != clone.uid
    assert len(clone.uid) > 32
    assert clone.id is None
    assert todo.filename != clone.filename
    assert clone.uid in clone.filename


@freeze_time('2017, 3, 20')
def test_todos_startable(tmpdir, runner, todo_factory, todos):
    todo_factory(summary='started', start=datetime(2017, 3, 15))
    todo_factory(summary='nostart')
    todo_factory(summary='unstarted', start=datetime(2017, 3, 24))

    todos = list(todos(startable=True))

    assert len(todos) == 2
    for todo in todos:
        assert 'unstarted' not in todo.summary


def test_filename_uid_colision(create, default_database, runner, todos):
    create('ABC.ics', 'SUMMARY:My UID is not ABC\n' 'UID:NOTABC\n')
    len(list(todos())) == 1

    todo = Todo(new=False)
    todo.uid = 'ABC'
    todo.list = next(default_database.lists())
    default_database.save(todo)

    len(list(todos())) == 2


def test_hide_cancelled(todos, todo_factory):
    todo_factory(status='CANCELLED')

    assert len(list(todos())) == 0
    assert len(list(todos(status='ANY'))) == 1


def test_illegal_start_suppression(create, default_database, todos):
    create(
        'test.ics', 'SUMMARY:Start doing stuff\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20170331T120000\n'
        'DTSTART;VALUE=DATE-TIME;TZID=CET:20170331T140000\n'
    )
    todo = next(todos())
    assert todo.start is None
    assert todo.due == datetime(2017, 3, 31, 12, tzinfo=tzoffset(None, 7200))


def test_default_status(create, todos):
    create('test.ics', 'SUMMARY:Finish all these status tests\n')
    todo = next(todos())
    assert todo.status == 'NEEDS-ACTION'


def test_nullify_field(default_database, todo_factory, todos):
    todo_factory(due=datetime.now())

    todo = next(todos(status='ANY'))
    assert todo.due is not None

    todo.due = None
    default_database.save(todo)

    todo = next(todos(status='ANY'))
    assert todo.due is None


def test_duplicate_list(tmpdir):
    tmpdir.join('personal1').mkdir()
    with tmpdir.join('personal1').join('displayname').open('w') as f:
        f.write('personal')

    tmpdir.join('personal2').mkdir()
    with tmpdir.join('personal2').join('displayname').open('w') as f:
        f.write('personal')

    with pytest.raises(AlreadyExists):
        Database(
            [tmpdir.join('personal1'),
             tmpdir.join('personal2')],
            tmpdir.join('cache.sqlite3'),
        )


def test_unreadable_ics(todo_factory, todos, tmpdir):
    """
    Test that we properly handle an unreadable ICS file

    In this case, it's a directory, which will
    fail even if you run the tests as root (you shouldn't!!), but the same
    codepath is followed for readonly files, etc.
    """
    tmpdir.join('default').join('fake.ics').mkdir()
    todo_factory()

    with patch('logging.Logger.exception') as mocked_exception:
        todos = list(todos())

    assert len(todos) == 1
    assert mocked_exception.call_count == 1


def test_cached_property_caching():
    class TestClass:
        i = 0

        @cached_property
        def a(self):
            TestClass.i += 1
            return TestClass.i

    obj = TestClass()
    assert obj.a == 1
    assert obj.a == 1
    assert obj.a == 1


def test_cached_property_overwriting():
    class TestClass:
        i = 0

        @cached_property
        def a(self):
            TestClass.i += 1
            return TestClass.i

    obj = TestClass()

    # Overriting will overwrite the cached_property:
    obj.a = 12
    assert obj.a == 12
    assert obj.a == 12

    obj.a += 1
    assert obj.a == 13


def test_cached_property_property():
    class TestClass:
        @cached_property
        def a(self):
            return 0

    assert TestClass.a.__class__ == cached_property



#!/usr/bin/env python3

from setuptools import setup

setup(
    name='todoman',
    description='A simple CalDav-based todo manager.',
    author='Hugo Osvaldo Barrera',
    author_email='hugo@barrera.io',
    url='https://github.com/pimutils/todoman',
    license='ISC',
    packages=['todoman'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'todo = todoman.cli:cli',
        ],
    },
    install_requires=[
        'atomicwrites',
        'click>=7.0,<8.0',
        'click-log>=0.2.1',
        'configobj',
        'humanize',
        'icalendar>=4.0.3',
        'parsedatetime',
        'python-dateutil',
        'pyxdg',
        'tabulate',
        'urwid',
    ],
    long_description=open('README.rst').read(),
    use_scm_version={
        'version_scheme': 'post-release',
        'write_to': 'todoman/version.py',
    },
    setup_requires=['setuptools_scm'],
    tests_require=open('requirements-dev.txt').readlines(),
    extras_require={
        'docs': open('requirements-docs.txt').readlines(),
        'repl': [
            'click-repl>=0.1.6',
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Utilities',
    ]
)



__all__ = [
    'pyicu_sensitive',
]

import os
from tempfile import TemporaryDirectory

import pytest


def is_fs_case_sensitive():
    with TemporaryDirectory() as tmpdir:
        os.mkdir(os.path.join(
            tmpdir,
            'casesensitivetest'
        ))
        try:
            os.mkdir(os.path.join(
                tmpdir,
                'casesensitiveTEST'
            ))
            return True
        except FileExistsError:
            return False


def is_pyicu_installed():
    try:
        import icu  # noqa: F401: This is an import to tests if it's installed.
    except ImportError:
        return False
    else:
        return True


fs_case_sensitive = pytest.mark.skipif(
    not is_fs_case_sensitive(),
    reason="This test cannot be run when the fs is not case sensitive.",
)


pyicu_sensitive = pytest.mark.skipif(
    is_pyicu_installed(),
    reason="This test cannot be run with pyicu installed.",
)



from datetime import date, datetime

import icalendar
import pytest
import pytz
from dateutil.tz import tzlocal
from freezegun import freeze_time

from todoman.model import Todo, VtodoWritter


def test_datetime_serialization(todo_factory, tmpdir):
    now = datetime(2017, 8, 31, 23, 49, 53, tzinfo=pytz.UTC)
    todo = todo_factory(created_at=now)
    filename = tmpdir.join('default').join(todo.filename)
    with open(str(filename)) as f:
        assert 'CREATED;VALUE=DATE-TIME:20170831T234953Z\n' in f.readlines()


def test_serialize_created_at(todo_factory):
    now = datetime.now(tz=pytz.UTC)
    todo = todo_factory(created_at=now)
    vtodo = VtodoWritter(todo).serialize()

    assert vtodo.get('created') is not None


def test_serialize_dtstart(todo_factory):
    now = datetime.now(tz=pytz.UTC)
    todo = todo_factory(start=now)
    vtodo = VtodoWritter(todo).serialize()

    assert vtodo.get('dtstart') is not None


def test_serializer_raises(todo_factory):
    todo = todo_factory()
    writter = VtodoWritter(todo)

    with pytest.raises(Exception):
        writter.serialize_field('nonexistant', 7)


def test_supported_fields_are_serializeable():
    supported_fields = set(Todo.ALL_SUPPORTED_FIELDS)
    serialized_fields = set(VtodoWritter.FIELD_MAP.keys())

    assert supported_fields == serialized_fields


def test_vtodo_serialization(todo_factory):
    """Test VTODO serialization: one field of each type."""
    description = 'A tea would be nice, thanks.'
    todo = todo_factory(
        categories=['tea', 'drinking', 'hot'],
        description=description,
        due=datetime(3000, 3, 21),
        start=date(3000, 3, 21),
        priority=7,
        status='IN-PROCESS',
        summary='Some tea',
        rrule='FREQ=MONTHLY',
    )
    writer = VtodoWritter(todo)
    vtodo = writer.serialize()

    assert (
        [str(c) for c in vtodo.get('categories').cats] ==
        ['tea', 'drinking', 'hot']
    )
    assert str(vtodo.get('description')) == description
    assert vtodo.get('priority') == 7
    assert vtodo.decoded('due') == datetime(3000, 3, 21, tzinfo=tzlocal())
    assert vtodo.decoded('dtstart') == date(3000, 3, 21)
    assert str(vtodo.get('status')) == 'IN-PROCESS'
    assert vtodo.get('rrule') == icalendar.vRecur.from_ical('FREQ=MONTHLY')


@freeze_time('2017-04-04 20:11:57')
def test_update_last_modified(todo_factory, todos, tmpdir):
    todo = todo_factory()
    assert todo.last_modified == datetime.now(tzlocal())


def test_sequence_increment(default_database, todo_factory, todos):
    todo = todo_factory()
    assert todo.sequence == 1

    default_database.save(todo)
    assert todo.sequence == 2

    # Relaod (and check the caching flow for the sequence)
    todo = next(todos())
    assert todo.sequence == 2


def test_normalize_datetime():
    writter = VtodoWritter(None)
    assert (
        writter.normalize_datetime(date(2017, 6, 17)) == date(2017, 6, 17)
    )
    assert (
        writter.normalize_datetime(datetime(2017, 6, 17)) == datetime(
            2017, 6, 17, tzinfo=tzlocal()
        )
    )
    assert (
        writter.normalize_datetime(datetime(2017, 6, 17, 12, 19)) == datetime(
            2017, 6, 17, 12, 19, tzinfo=tzlocal()
        )
    )
    assert (
        writter.normalize_datetime(
            datetime(2017, 6, 17, 12, tzinfo=tzlocal())
        ) == datetime(2017, 6, 17, 12, tzinfo=tzlocal())
    )
